"""
Module for handling API calls to the [AlgoBulls](https://www.algobulls.com) backend.
"""

from json import JSONDecodeError

import requests

from .exceptions import AlgoBullsAPIBaseException, AlgoBullsAPIUnauthorizedError, AlgoBullsAPIResourceNotFoundError, AlgoBullsAPIBadRequest, AlgoBullsAPIInternalServerErrorException
from ..constants import StrategyType


class AlgoBullsAPI:
    """
    AlgoBulls API
    """
    # SERVER_ENDPOINT = 'https://api.algobulls.com/'
    SERVER_ENDPOINT = 'http://127.0.0.1:8000/'

    def __init__(self):
        """
        Init method that is used while creating an object of this class
        """
        self.headers = None

    def set_access_token(self, access_token: str):
        """
        Sets access token to the header attribute, which is needed for APIs requiring authorization
        Package for interacting with AlgoBulls Algorithmic Trading Platform (https://www.algobulls.com)

        Args:
            access_token: Access token generated by logging to the URL given by the `get_authorization_url()` method
        """
        self.headers = {
            'Authorization': f'{access_token}'
        }

    def _send_request(self, method: str = 'get', endpoint: str = '', base_url: str = SERVER_ENDPOINT, params: [str, dict] = None, json_data: [str, dict] = None, requires_authorization: bool = True) -> dict:
        """
        Send the request to the platform
        Args:
            method: get
            endpoint: endpoint url
            base_url: base url
            params: parameters
            json_data: json data as body
            requires_authorization: True or False

        Returns:
            request status
        """
        url = f'{base_url}{endpoint}'
        headers = self.headers if requires_authorization else None
        response = requests.request(method=method, headers=headers, url=url, params=params, json=json_data)

        try:
            response_json = response.json()
        except JSONDecodeError:
            response_json = str(response)

        if response.status_code == 200:
            response_json = response.json()
            return response_json
        elif response.status_code == 400:
            raise AlgoBullsAPIBadRequest(f'Bad Request --> Method: {method} | URL: {url} | Response: {response_json}')
        elif response.status_code == 401:
            raise AlgoBullsAPIUnauthorizedError(f'Unauthorized Error --> Method: {method} | URL: {url} | Response: {response_json}')
        elif response.status_code == 404:
            raise AlgoBullsAPIResourceNotFoundError(f'API Not Found --> Method: {method} | URL: {url} | Response: {response_json}')
        elif response.status_code == 500:
            raise AlgoBullsAPIInternalServerErrorException(f'Internal Server Error --> Method: {method} | URL: {url} | Response: {response_json}')
        else:
            response.raw.decode_content = True
            raise AlgoBullsAPIBaseException(f'Unknown non-200 status code --> Method: {method} | URL: {url} | Response: {response_json} | Response code: {response.status_code}')

    def create_strategy(self, strategy_name: str, strategy_details: str, abc_version: str) -> dict:
        """
        Create a new strategy for the user on the AlgoBulls platform.

        Args:
            strategy_name: name of the strategy
            strategy_details: Python code of the strategy
            abc_version: value of one of the enums available under [AlgoBullsEngineVersion]()

        Returns:
            JSON Response received from AlgoBulls platform after (attempt to) creating a new strategy.

        Warning:
            For every user, the `strategy_name` should be unique. You cannot create multiple strategies with the same name.

        Info: ENDPOINT
            `POST` v2/user/strategy/build/python
        """
        json_data = {'strategyName': strategy_name, 'strategyDetails': strategy_details, 'abcVersion': abc_version}
        endpoint = f'v2/user/strategy/build/python'
        response = self._send_request(endpoint=endpoint, method='post', json_data=json_data)
        return response

    def update_strategy(self, strategy_name: str, strategy_details: str, abc_version: str) -> dict:
        """
        Update an already existing strategy on the AlgoBulls platform

        Args:
            strategy_name: name of the strategy
            strategy_details: Python code of the strategy
            abc_version: value of one of the enums available under `AlgoBullsEngineVersion`

        Returns:
            JSON Response received from AlgoBulls platform after (attempt to) updating an existing strategy.

        Info: ENDPOINT
            PUT v2/user/strategy/build/python
        """
        json_data = {'strategyName': strategy_name, 'strategyDetails': strategy_details, 'abcVersion': abc_version}
        endpoint = f'v2/user/strategy/build/python'
        response = self._send_request(endpoint=endpoint, method='put', json_data=json_data)
        return response

    def get_all_strategies(self) -> dict:
        """
        Get all the Python strategies created by the user on the AlgoBulls platform

        Returns:
            JSON Response received from AlgoBulls platform with list of all the created strategies.

        Info: ENDPOINT
            `OPTIONS` v2/user/strategy/build/python
        """
        endpoint = f'v2/user/strategy/build/python'
        response = self._send_request(endpoint=endpoint, method='options')
        return response

    def get_strategy_details(self, strategy_code: str) -> dict:
        """
        Get strategy details for

        Arguments:
            strategy_code: unique code of strategy, which is received while creating the strategy or

        Return:
            JSON

        Info: ENDPOINT
            `GET` v2/user/strategy/build/python
        """
        params = {'strategyCode': strategy_code}
        endpoint = f'v2/user/strategy/build/python'
        response = self._send_request(endpoint=endpoint, params=params)
        return response

    def search_instrument(self, instrument: str) -> dict:
        """

        Args:
            instrument: instrument key

        Returns:
            JSON Response


        Info: ENDPOINT
            `GET` v2/instrument/search
        """
        params = {'instrument': instrument}
        endpoint = f'v2/instrument/search'
        response = self._send_request(endpoint=endpoint, params=params, requires_authorization=False)
        return response

    def set_strategy_config(self, strategy_code: str, strategy_config: dict) -> dict:
        """

        Args:
            strategy_code: strategy code
            strategy_config: strategy configuration

        Returns:

        Info: ENDPOINT
           PATCH v1/customer_strategy_tweak
        """
        params = {'strategyCode': strategy_code, 'strategyType': StrategyType.PYTHON.value}
        json_data = strategy_config
        endpoint = f'v1/customer_strategy_tweak'
        response = self._send_request(method='patch', endpoint=endpoint, params=params, json_data=json_data)
        return response

    def start_strategy_algotrading(self, strategy_code: str, trading_type: str, broker: str = '') -> dict:
        """
        Submit Backtesting / Paper Trading / Real Trading job for strategy with code strategy_code & return the job ID.

        Info: ENDPOINT
            `POST` v1/customer_strategy_algotrading
        """
        params = {'strategyCode': strategy_code, 'strategyType': StrategyType.PYTHON.value, 'tradingType': trading_type.value, 'broker': broker}
        endpoint = f'v1/customer_strategy_algotrading'
        json_data = {'action': 'start'}
        response = self._send_request(method='post', endpoint=endpoint, params=params, json_data=json_data)
        return response

    def stop_strategy_algotrading(self, strategy_code: str, trading_type: str, broker: str = '') -> dict:
        """
        Stop Backtesting / Paper Trading / Real Trading job for strategy with code strategy_code & return the job ID.

        Info: ENDPOINT
            `POST` v1/customer_strategy_algotrading
        """
        params = {'strategyCode': strategy_code, 'strategyType': StrategyType.PYTHON.value, 'tradingType': trading_type.value, 'broker': broker}
        endpoint = f'v1/customer_strategy_algotrading'
        json_data = {'action': 'stop'}
        response = self._send_request(method='post', endpoint=endpoint, params=params, json_data=json_data)
        return response

    def get_job_status(self, strategy_code: str, trading_type: str, broker: str = '') -> dict:
        """


        Get status for a BACKTESTING/PAPERTESTING/REALTRADING Job

        Args:
            strategy_code: Strategy code
            trading_type: Trading type
            broker: Name of the broker

        Returns:
            Job status

        Info: ENDPOINT
            `GET` v1/customer_strategy_algotrading
        """
        params = {'strategyCode': strategy_code, 'strategyType': StrategyType.PYTHON.value, 'tradingType': trading_type.value, 'broker': broker}
        endpoint = f'v1/customer_strategy_algotrading'
        response = self._send_request(endpoint=endpoint, params=params)
        return response

    def get_reports(self, strategy_code: str, trading_type: str, report_type: str, broker: str = '') -> dict:
        """
        Get reports for a BACKTESTING/PAPERTESTING/REALTRADING Job

        Args:
            strategy_code: Strategy code
            trading_type: Value of TradingType Enum
            report_type: Value of TradingReportType Enum
            broker: Name of the broker

        Returns:
            Report data

        Info: ENDPOINT
            `GET` v1/customer_strategy_algotrading_reports
        """
        params = {'strategyCode': strategy_code, 'strategyType': StrategyType.PYTHON.value, 'tradingType': trading_type.value, 'reportType': report_type.value, 'broker': broker}
        endpoint = f'v1/customer_strategy_algotrading_reports'
        response = self._send_request(endpoint=endpoint, params=params)
        return response
